/*********************	Written by Christian Nordahl  ***********************/
/*
 *	This program simulates the DVB-RCS encoding-decoding system, for UNIX.
 *
 *	It uses the CRSC codes in DVB-RCS system model, with two component CRSC encoders
 *	( Circular Recursive Systematic Convolutional).
 *
 *	Random information bits are modulated into +1/-1, and transmitted through an AWGN channel.
 *
 *	The Max-Log-MAP algorithm with the approximation ln(e^x+e^y+e^z+e^w) ≈ max(x,y,z,w) is used.
 *
 *	To set the number of iterations, the globle variable "DECITER" should be changed.
 *	To set the frame length, the global variable "N" should be changed.
 *	To set the code rate, the global variable "R" should be changed.
 *	To simulate a different range of Eb/N0, change the global variable LOEBNO and HIEBNO,
 *	as well as the step size of Eb/No increment.
 */

/* #include <malloc.h> This needs to be uncommented for the PC version */

#include <stdlib.h>
#include <math.h>
#include <stdio.h>
#include <time.h>
#include "turbo_dvbrcs.h"

#define K 4					/* constraint length */
#define NUMSTATES 8			/* 2^(K - 1) */
#define SYMBOLS	4			/* The number of symbols. 4 for duo-binary */

void gen01dat(int data_len, int *out_array);
void permutation(int *permute_index);
void deci2bin(int d, int size, int *b);
void turbo_encd(int g[4][K], int M, int *permute_index, int *in_array, int *turbo_en_out);
void addnoise(float eb_ovr_n0, int encd_block_len, float rate, int *in_array, float *out_array);
void de_multiplex(int msg_length, int M, int *permute_index, float *in_array, float *sys_array,
				  float *pinfo_array, float *parity1_array, float *parity2_array);
void maxlogmap(float *rec_s, float *rec_p, float *L_a, float *L_u, float *Alpha0, float *BetaN, int nextstate[NUMSTATES][SYMBOLS], int nextoutput[NUMSTATES][NUMSTATES][2],
			   int prestate[NUMSTATES][SYMBOLS], int preoutput[NUMSTATES][NUMSTATES][2], int qpsk_input[SYMBOLS][2]);
void trellis(int g[4][K], int nextstate[NUMSTATES][SYMBOLS], int nextoutput[NUMSTATES][NUMSTATES][2],
			 int prestate[NUMSTATES][SYMBOLS], int preoutput[NUMSTATES][NUMSTATES][2]);
float max3(float x, float y, float z);

Results *dvbrcsSimulation()
{
	int i, j, ll, iter, t, start, totaltime;   /* loop variables */
	int encd_block_len,total_len;   /* the length of the encoded block depending on the code rate */
	int M,msg_length,channel_len;   /* parameters of encoded block length */
	int a,dec_iter,snrIter;         /* various helper variables */
	
	/* Tables used to represent the trellis, used in decoding by maxlogmap() */
	int nextstate[NUMSTATES][SYMBOLS], prestate[NUMSTATES][SYMBOLS]; 
	int nextoutput[NUMSTATES][NUMSTATES][2], preoutput[NUMSTATES][NUMSTATES][2];
	
	/* Variables used for error control and validation of various elements of the implementation, returned in the Result-struct */
	int snrIterations,onescount,zeroescount;
	int *errorPositions,*onesc,*zeroesc,*frames;
	double e_ber,f_ber;
	double *berRet,*ferRet;
	
	int *permute_index;                     /* permutation table */
	int *onezer;					/* random systematic bits */
	int *encoded;                   /* original, encoded array */
	float *splusn;                  /* noisy data array */
	float *sys_array;               /* received systematic array as input for decoder one */
	float *pinfo_array;             /* received systematic array, interleaved as input for decoder two */
	float *parity1_array;           /* received parity1 array for decoder one */
	float *parity2_array;           /* received parity2 array for decoder two */
	float *L_a;                     /* deinterleaved or interleaved extrinsic info */
	float *L_e_1;                   /* decoder two output, then deinterleaved as input for decoder one */
	float *L_e_2;                   /* decoder one output, then interleaved as input for decoder two */
	float *L_u;                     /* log_likelihood ratios */
	float *L_all;                   /* final soft output after iterations */
	int *turbout;                   /* estimated bit of decoder */
	float yx[N][SYMBOLS],L_a_p,L_e_2_p;   /* temp values for calculating extrinsic info, interleaving, deinterleaving */
	float eb_ovr_n0,rate,L_c,bit_errors,temp_bit_errors,frame_errors,e_threshold;   /* various statistics */
	
	float Ad1[NUMSTATES],Bd1[NUMSTATES],Ad2[NUMSTATES],Bd2[NUMSTATES];   /* circular trellis state values */
	float *Alpha0,*BetaN;
	
	static int g[4][K] = {
		{1, 1,0,1},			/* Recursion: 15 octally */
		{1, 0,1,1},			/* Y-redundance: 13 octally */
		{1, 0,0,1},			/* W-redundance: 11 octally */
		{0, 1,1,0}};		/* Helper mask for updating the shift register with bit B: 06 octally */
	
	static int qpsk_input[SYMBOLS][2] = { 
		{ 1, 1},
		{ 1,-1},
		{-1, 1},
		{-1,-1}};
	
	FILE *fp;
		
#if R == 1
	rate = 1.0/3.0;
	M = N;
	encd_block_len = 2*N + M;
#endif
	
#if R == 2
	rate = 2.0/5.0;
	M = N/2;
	encd_block_len = 2*N + M;
#endif
	
#if R == 3
	rate = 1.0/2.0;
	M = N;
	encd_block_len = N + M;
#endif
	
#if R == 4
	rate = 2.0/3.0;
	M = N/2;
	encd_block_len = N + M;
#endif
	
#if R == 5
	rate = 3.0/4.0;
	if (N%3 == 0) M = N/3;
	else if (N%3 == 1) M = (N-4)/3 + 2;
	else if (N%3 == 2) M = (N-8)/3 + 3;
	encd_block_len = N + M;
#endif
	
#if R == 6
	rate = 4.0/5.0;
	M = N/4;
	encd_block_len = N + M;
#endif
	
#if R == 7
	rate = 6.0/7.0;
	if (N%3 == 0) M = N/6;
	else if (N%3 == 1) M = (N-4)/6 + 1;
	else if (N%3 == 2) M = (N-8)/6 + 2;
	encd_block_len = N + M;
#endif
	
	msg_length = 2*N;
	total_len = 3*N;
	channel_len = 2*encd_block_len;

	snrIterations = floor( (HIEBN0 - LOEBN0) / EBN0STEP ) + 1;
	
	/* Allocating memory for the result arrays to be returned in the Result struct */
	errorPositions = malloc( snrIterations * msg_length * sizeof(int) );
	if (errorPositions == NULL) {
		printf("\n turbo_dvbrcs.c:  error allocating temporary result table, aborting!");
		exit(1);
	}
	
	for (t = 0; t < snrIterations * msg_length; t++){     /* initializing the array */
		*(errorPositions + t) = 0;
	}
	
	frames = malloc( snrIterations * sizeof(int) );
	if (frames == NULL) {
		printf("\n turbo_dvbrcs.c:  error allocating temporary result table, aborting!");
		exit(1);
	}
	
	onesc = malloc( snrIterations * sizeof(int) );
	if (onesc == NULL) {
		printf("\n turbo_dvbrcs.c:  error allocating temporary result table, aborting!");
		exit(1);
	}
	
	zeroesc = malloc( snrIterations * sizeof(int) );
	if (zeroesc == NULL) {
		printf("\n turbo_dvbrcs.c:  error allocating temporary result table, aborting!");
		exit(1);
	}
	
	berRet = malloc( snrIterations * sizeof(double) );
	if (berRet == NULL) {
		printf("\n turbo_dvbrcs.c:  error allocating temporary result table, aborting!");
		exit(1);
	}
	
	ferRet = malloc( snrIterations * sizeof(double) );
	if (ferRet == NULL) {
		printf("\n turbo_dvbrcs.c:  error allocating temporary result table, aborting!");
		exit(1);
	}
	
	
	/* generate the code trellis arrays */ 
	trellis(g, nextstate, nextoutput, prestate, preoutput);
	
	fp = fopen("brage_turbo_dvbrcs_data3","a");
	if ((fp = fopen("brage_turbo_dvbrcs_data3","a"))==NULL){
		printf("cannot open file\n");
		exit (0);
	}

	printf("\ng1 = %d%d%d%d", g[0][0], g[0][1], g[0][2], g[0][3]);
	printf("\ng2 = %d%d%d%d", g[1][0], g[1][1], g[1][2], g[1][3]);
	printf("\ng3 = %d%d%d%d", g[2][0], g[2][1], g[2][2], g[2][3]);
	printf("\ng4 = %d%d%d%d\n", g[3][0], g[3][1], g[3][2], g[3][3]);
	fprintf(fp,"\ng1 = %d%d%d%d", g[0][0], g[0][1], g[0][2], g[0][3]);
	fprintf(fp,"\ng2 = %d%d%d%d", g[1][0], g[1][1], g[1][2], g[1][3]);
	fprintf(fp,"\ng3 = %d%d%d%d", g[2][0], g[2][1], g[2][2], g[2][3]);
	fprintf(fp,"\ng4 = %d%d%d%d\n", g[3][0], g[3][1], g[3][2], g[3][3]);
	
	printf("\nTURBO DVBRCS\nK = %d   Frame size N = %d  M = %d  code rate = %f\n", K, N, M, rate);
	fprintf(fp,"\nTURBO DVBRCS\nK = %d   Frame size N = %d  M = %d  code rate = %f\n", K, N, M, rate);
	
	fclose(fp);
	
	snrIter = 0;
	totaltime = time(NULL);
	
	for (eb_ovr_n0 = LOEBN0; eb_ovr_n0 <= HIEBN0; eb_ovr_n0 += EBN0STEP) {
		
		start = time(NULL);
		
		a = 1;
		L_c = 4 * a * rate * pow(10,eb_ovr_n0/10); 
		bit_errors = 0.0;
		frame_errors = 0.0;
		e_ber = 0.0;
		f_ber = 0.0;
		iter = 0;
		onescount = 0;
		zeroescount = 0;
		e_threshold = MINERRORS;
		srand(time(NULL));
		
		while  ( ((iter < MAXFRAMES) && (bit_errors < e_threshold)) || (iter < MINFRAMES) )  { /*  ( ((iter<MAXFRAMES) && (bit_errors < e_threshold)) || (iter < MINFRAMES) )  */
			permute_index = malloc( N * sizeof(int) );
			if (permute_index == NULL) {
				printf("\n turbo_dvbrcs.c:  error allocating interleaver table, aborting!");
				exit(1);
			}
			
			onezer = malloc( msg_length * sizeof(int) );
			if (onezer == NULL) {
				printf("\n turbo_dvbrcs.c:  error allocating onezer array, aborting!");
				exit(1);
			}
			
			encoded = malloc( channel_len * sizeof(int) );
			if (encoded == NULL) {
				printf("\n turbo_dvbrcs.c:  error allocating encoded array, aborting!");
				exit(1);
			}
			
			splusn = malloc( channel_len * sizeof(float) );
			if (splusn == NULL) {
				printf("\n turbo_dvbrcs.c:  error allocating splusn array, aborting!");
				exit(1);
			}
			
			sys_array = malloc( msg_length * sizeof(float) );
			if (sys_array == NULL) {
				printf("\n turbo_dvbrcs.c:  error allocating received symbol sys array, aborting!");
				exit(1);
			}
			
			pinfo_array = malloc( msg_length * sizeof(float) );
			if (pinfo_array == NULL) {
				printf("\n turbo_dvbrcs.c:  error allocating received symbol interleaved sys array, aborting!");
				exit(1);
			}
			
			parity1_array = malloc( msg_length * sizeof(float) );
			if (parity1_array == NULL) {
				printf("\n turbo_dvbrcs.c:  error allocating received symbol parity1 array, aborting!");
				exit(1);
			}
			
			parity2_array = malloc( msg_length * sizeof(float) );
			if (parity2_array == NULL) {
				printf("\n turbo_dvbrcs.c:  error allocating received symbol parity2 array, aborting!");
				exit(1);
			}
			
			Alpha0 = malloc( NUMSTATES * sizeof(float) );
			if (Alpha0 == NULL) {
				printf("\n turbo_dvbrcs.c:  error allocating the values of circular trellis state, aborting!");
				exit(1);
			}
			
			BetaN = malloc( NUMSTATES * sizeof(float) );
			if (BetaN == NULL) {
				printf("\n turbo_dvbrcs.c:  error allocating the values of circular trellis state, aborting!");
				exit(1);
			}
			
			L_a = malloc( total_len * sizeof(float) );
			if (L_a == NULL) {
				printf("\n turbo_dvbrcs.c:  error allocating L_a array, aborting!");
				exit(1);
			}
			
			L_e_1 = malloc( total_len * sizeof(float) );
			if (L_e_1 == NULL) {
				printf("\n turbo_dvbrcs.c:  error allocating L_e_1 array, aborting!");
				exit(1);
			}
			
			L_e_2 = malloc( total_len * sizeof(float) );
			if (L_e_2 == NULL) {
				printf("\n turbo_dvbrcs.c:  error allocating L_e_2 array, aborting!");
				exit(1);
			}
			
			L_u = malloc( total_len * sizeof(float) );
			if (L_u == NULL) {
				printf("\n turbo_dvbrcs.c:  error allocating L_u array, aborting!");
				exit(1);
			}
			
			L_all = malloc( total_len * sizeof(float) );
			if (L_all == NULL) {
				printf("\n turbo_dvbrcs.c:  error allocating L_all array, aborting!");
				exit(1);
			}
			
			turbout = malloc( msg_length * sizeof(int) );
			if (turbout == NULL) {
				printf("\n turbo_dvbrcs.c:  error allocating turbo decoding out array, aborting!");
				exit(1);
			}
			
			iter += 1;
			
			gen01dat(msg_length, onezer);

			permutation(permute_index);
			
			turbo_encd(g, M, permute_index, onezer, encoded);
			
			addnoise(eb_ovr_n0, encd_block_len, rate, encoded, splusn);
			
			de_multiplex(msg_length, M, permute_index, splusn, sys_array, pinfo_array, parity1_array, parity2_array);
			
			for (j=0; j<msg_length; j++){
				*(sys_array+j) = 0.5*L_c**(sys_array+j);
				*(pinfo_array+j) = 0.5*L_c**(pinfo_array+j);
				*(parity1_array+j) = 0.5*L_c**(parity1_array+j);
				*(parity2_array+j) = 0.5*L_c**(parity2_array+j);
			}
			
			for (j=0; j<total_len; j++){           /* initial extrinsic info and log_likelihood ratio */
				*(L_a+j) = 0.0;
				*(L_e_1+j) = 0.0;
				*(L_e_2+j) = 0.0;
				*(L_u+j) = 0.0;
				*(L_all+j) = 0.0;
			}
			
			for (i=0; i<NUMSTATES; i++){           /* initial circular trellis values */
				Ad1[i] = 0.0;
				Bd1[i] = 0.0;
				Ad2[i] = 0.0;
				Bd2[i] = 0.0;
			}
			
			for (dec_iter=0; dec_iter<DECITER; dec_iter++){
				for (i=0; i<3; i++){                 /* deinterleave the extrinsic info for Decoder one */
					for (j=0; j<N; j++)                /* deinterleaving level two */
						*(L_a + *(permute_index+j) + i*N) = *(L_e_1 + j + i*N);
				}
				
				for (j=0; j<N; j++){                 /* deinterleaving level one */
					if (j%2==0){
						L_a_p = *(L_a+j);
						*(L_a+j) = *(L_a+j+N);
						*(L_a+j+N) = L_a_p;
					}
				}
				
				for (i=0; i<NUMSTATES; i++){
					*(Alpha0 + i) = Ad1[i];
					*(BetaN + i) = Bd1[i];
				}
				
				maxlogmap(sys_array, parity1_array, L_a, L_u, Alpha0, BetaN, nextstate, nextoutput, 
						  prestate, preoutput,qpsk_input);
				
				for (i=0; i<NUMSTATES; i++){
					Ad1[i] = *(Alpha0 + i);
					Bd1[i] = *(BetaN + i);
				}
				
				for (j=0; j<N; j++){                 /* get extrinsic info for Decoder two */
					for (i=0; i<4; i++){
						yx[j][i] = *(sys_array + 2*j)*qpsk_input[i][0] + *(sys_array + 2*j + 1)*qpsk_input[i][1];
					}
					
					for (i=0; i<3; i++)
						*(L_e_2+j+i*N) = *(L_u+j+i*N) - yx[j][i+1] + yx[j][0] - *(L_a+j+i*N);
				}
				
				for (j=0; j<N; j++){                 /* interleaving extrinsic info for Decoder two, level one */
					if (j%2==0){
						L_e_2_p = *(L_e_2+j);
						*(L_e_2+j) = *(L_e_2+j+N);
						*(L_e_2+j+N) = L_e_2_p;
					}
				}
				
				for (i=0; i<3; i++){                 /* interleaving level two */
					for (j=0; j<N; j++)
						*(L_a + j + i*N) = *(L_e_2 + i*N + *(permute_index+j));
				}
				
				for (i=0; i<NUMSTATES; i++){
					*(Alpha0 + i) = Ad2[i];
					*(BetaN + i) = Bd2[i];
				}
				
				maxlogmap(pinfo_array, parity2_array, L_a, L_u, Alpha0, BetaN, nextstate, nextoutput, 
						  prestate, preoutput,qpsk_input);  /* Decoder two */
				
				for (i=0; i<NUMSTATES; i++){
					Ad2[i] = *(Alpha0 + i);
					Bd2[i] = *(BetaN + i);
				}
				
				for (j=0; j<N; j++){                 /* get extrinsic info for Decoder one */
					for (i=0; i<4; i++){
						yx[j][i] = *(pinfo_array + 2*j)*qpsk_input[i][0] + *(pinfo_array + 2*j + 1)*qpsk_input[i][1];
					}
					for (i=0; i<3; i++)
						*(L_e_1+j+i*N) = *(L_u+j+i*N) - yx[j][i+1] + yx[j][0] - *(L_a+j+i*N);
				}
			}  /* end dec_iter-loop */

			
			for (i=0; i<3; i++){						/* deinterleave the log_likelihood ratios */
				for (j=0; j<N; j++)
					*(L_all + *(permute_index+j) + i*N) = *(L_u + j + i*N);
			}
			
			/* estimation of the decoded output, including demodulation from QPSK */
			for (j=0; j<N; j++){
				if ( max3(*(L_all+j),*(L_all+j+N),*(L_all+j+2*N)) > 0.0 ){
					if ( *(L_all+j)==max3(*(L_all+j),*(L_all+j+N),*(L_all+j+2*N)) ){
						if(j%2 == 0){					/* deinterleaver for j mod 2 = 0, invert A and B */
							*(turbout+2*j) = 1;
							*(turbout+2*j+1) = 0;
						}
						else {
							*(turbout+2*j) = 0;
							*(turbout+2*j+1) = 1;
						}
					}
					
					else if ( *(L_all+j+N)==max3(*(L_all+j),*(L_all+j+N),*(L_all+j+2*N)) ){
						if(j%2 == 0){					/* deinterleaver for j mod 2 = 0, invert A and B */
							*(turbout+2*j) = 0;
							*(turbout+2*j+1) = 1;
						}
						else {
							*(turbout+2*j) = 1;
							*(turbout+2*j+1) = 0;
						}
					}
					
					else if ( *(L_all+j+2*N)==max3(*(L_all+j),*(L_all+j+N),*(L_all+j+2*N)) ){
						*(turbout+2*j) = 1;
						*(turbout+2*j+1) = 1;
					}
				}
				
				else {
					*(turbout+2*j) = 0;
					*(turbout+2*j+1) = 0;
				}
			} /* end j for-estimating loop */
			
			temp_bit_errors = bit_errors;
			for (t = 0; t < msg_length; t++){     /* count the bit errors */
				if (*(onezer + t) != *(turbout + t)){
					bit_errors = bit_errors + 1;
					*(errorPositions + t + (snrIter * msg_length)) = *(errorPositions + t + (snrIter * msg_length)) + 1;
				}
				if (*(onezer + t) == 1){
					onescount= onescount + 1;
				}
				else {
					zeroescount= zeroescount + 1;
				}
			} /* end t count bit errors-loop */
			
			if ((bit_errors - temp_bit_errors) > 0) frame_errors = frame_errors + 1;   /* count the frame errors */
			
			/* free memory */
			free(permute_index);
			free(onezer);
			free(encoded);
			free(splusn);
			free(sys_array);
			free(pinfo_array);
			free(parity1_array);
			free(parity2_array);
			free(Alpha0);
			free(BetaN);
			free(L_a);
			free(L_e_1);
			free(L_e_2);
			free(L_u);
			free(L_all);
			free(turbout);
		} /* end of while loop */
		
		e_ber = bit_errors / (msg_length * iter);
		f_ber = frame_errors / iter;
		
		/* storing the values from the current snr-iteration */
		*(onesc + snrIter) = onescount;
		*(zeroesc + snrIter) = zeroescount;
		*(frames + snrIter) = iter;
		
		*(berRet + snrIter) = e_ber;
		*(ferRet + snrIter) = f_ber;
		
		
		fp = fopen("brage_turbo_dvbrcs_data3","a");
		if ((fp = fopen("brage_turbo_dvbrcs_data3","a"))==NULL){
			printf("cannot open file\n");
			exit (0);
		}
		
		printf("\nChannel reliability Lc = %f   BER threshold = %1.0f", L_c, e_threshold);
		fprintf(fp,"\nChannel reliability Lc = %f   BER threshold = %1.0f", L_c, e_threshold);
		
		printf("\nThe elapsed time was %d seconds for %d frames and %d decoding iterations",
			   time(NULL) - start, iter, dec_iter);
		fprintf(fp,"\nThe elapsed time was %d seconds for %d frames and %d decoding iterations",
				time(NULL) - start, iter, dec_iter);
		
		printf("\nAt %1.4fdB Eb/No, ", eb_ovr_n0);
		fprintf(fp,"\nAt %1.4fdB Eb/No, ", eb_ovr_n0);
		
		printf("the e_ber was %1.4e the f_ber was %1.4e \n", e_ber, f_ber);
		fprintf(fp,"the e_ber was %1.4e the f_ber was %1.4e\n", e_ber, f_ber);
		
		fclose (fp);
		
		snrIter += 1;
	} /* end of eb_ovr_n0 loop */
	
	
	fp = fopen("brage_turbo_dvbrcs_data3","a");
	if ((fp = fopen("brage_turbo_dvbrcs_data3","a"))==NULL){
		printf("cannot open file\n");
		exit (0);
	}
	printf("\nThe total simulation time was %d seconds for all snr-levels",
		   (time(NULL) - totaltime));
	fprintf(fp,"\nThe total simulation time was %d seconds for all snr-levels",
			(time(NULL) - totaltime));
			
	fclose (fp);
	
	Results *res;
	res = malloc( sizeof(Results) );
	if (res == NULL) {
		printf("\n turbo_dvbrcs.c: error allocating result struct, aborting!");
		exit(1);
	}
	
	/* linking result tables */
	res->errorPos = errorPositions;
	res->frames = frames;
	res->onecheck = onesc;
	res->zerocheck = zeroesc;
	res->ber = berRet;
	res->fer = ferRet;
	
	/* linking result scalars */
	res->blockLength = N;
	res->rate = R;
	res->snrIters = snrIterations;
	
	return res;
}

float max3(float x, float y, float z)
{
	float maxmum;
	
	maxmum = x>y?x:y;
	maxmum = maxmum>z?maxmum:z;
	
	return(maxmum);
}

/*******************/
/* 0/1 Generation  */
/*******************/
void gen01dat( int data_len, int *out_array )
{	
	int t;							/* time */
	for (t = 0; t < data_len; t++)	/* generate random data, write to output array */
		*( out_array + t ) = (int)( rand() / (RAND_MAX / 2) > 0.5 );
}

/***************************/
/* PCCC Turbo code encoder */
/***************************/
void nbrsc_encd(int g[4][K], int *in_array, int *y_array, int *w_array, int s0[K-1], int sN[K-1]);
void deci2bin(int d, int size, int *b);
int bin2deci(int *b, int size);
int start_end_state(int sNdeci);

void turbo_encd(int g[4][K], int M, int *permute_index, int *in_array, int *turbo_en_out)
{
	int i,j,temp_data;                 /* loop variables, temporary values */
	int y[2][N];                       /* matrix for storing the temporary output */
	int *pinfo_array;                  /* interleaved input message data */
	int *Y1out_array,*W1out_array;     /* encoder C1 output data */
	int *Y2out_array,*W2out_array;     /* encoder C2 output data */
	int s0deci,sNdeci;				   /* state numbers in decimal */
	int s0[K-1],sN[K-1];               /* state numbers in binary */
	
	pinfo_array = malloc( (2*N) * sizeof(int) );
	if (pinfo_array == NULL) {
		printf("\n turbo_encd:  error allocating interleaved info array, aborting!");
		exit(1);
	}
	
	Y1out_array = malloc( N * sizeof(int) );
	if (Y1out_array == NULL) {
		printf("\n turbo_encd:  error allocating Y array of encoder one, aborting!");
		exit(1);
	}
	
	W1out_array = malloc( N * sizeof(int) );
	if (W1out_array == NULL) {
		printf("\n turbo_encd:  error allocating W array of encoder one, aborting!");
		exit(1);
	}
	
	Y2out_array = malloc( N * sizeof(int) );
	if (Y2out_array == NULL) {
		printf("\n turbo_encd:  error allocating Y array of encoder two, aborting!");
		exit(1);
	}
	
	W2out_array = malloc( N * sizeof(int) );
	if (W2out_array == NULL) {
		printf("\n turbo_encd:  error allocating W array of encoder two, aborting!");
		exit(1);
	}
	
	s0deci = 0;
	deci2bin(s0deci, K-1, s0);
	
	nbrsc_encd(g, in_array, Y1out_array, W1out_array, s0, sN);  /* trial coding to get the initial state */
	
	sNdeci = bin2deci(sN, K-1);
	s0deci = start_end_state(sNdeci);
	deci2bin(s0deci, K-1, s0);
	
	nbrsc_encd(g, in_array, Y1out_array, W1out_array, s0, sN);  /* encoder C1 output */
	
	/* make a matrix with first and second row corresponding to info sequence */
	/* third and fifth row corresponding to RSC #1 s check/parity bits */
	/* fourth and sixth row corresponding to RSC #2 s check/parity bits */
	
	for (j=0; j<2*N; j++){         /* get the systematic information array */
		*(pinfo_array + j) = *(in_array + j);
		*(turbo_en_out + j) = *(in_array + j);
	}
	
	for (j=0; j<N; j++){           /* turbo code permutation Level one */
		if  (j%2 == 0){
			temp_data = *(pinfo_array + 2*j);
			*(pinfo_array + 2*j) = *(pinfo_array + 2*j + 1);
			*(pinfo_array + 2*j + 1) = temp_data;
		}
	}
	
	for (i=0; i<2; i++){           /* turbo code permutation Level two */
		for (j=0; j<N; j++)
			y[i][j] = *(pinfo_array + 2*j + i);
	}
	
	for (i=0; i<2; i++){
		for (j=0; j<N; j++)
			*(pinfo_array + 2*j + i) = y[i][*(permute_index + j)];
	}
	
	s0deci = 0;
	deci2bin(s0deci, K-1, s0);
	
	nbrsc_encd(g, pinfo_array, Y2out_array, W2out_array, s0, sN);  /* Test the trellis to get the initial state */
	
	sNdeci = bin2deci(sN, K-1);
	s0deci = start_end_state(sNdeci);
	deci2bin(s0deci, K-1, s0);
	
	nbrsc_encd(g, pinfo_array, Y2out_array, W2out_array, s0, sN);  /* encoder C2 output */
	
#if R == 1              /* code rate = 1/3 */
	for (j=0; j<N; j++){
		*(turbo_en_out + 2*j + 2*N) = *(Y1out_array + j);
		*(turbo_en_out + 2*j + 1 + 2*N) = *(Y2out_array + j);
		*(turbo_en_out + 2*j + 4*N) = *(W1out_array + j);
		*(turbo_en_out + 2*j + 1 + 4*N) = *(W2out_array + j);
	}
#endif
	
#if R == 2             /* code rate = 2/5 */
	for (j=0; j<N; j++){
		*(turbo_en_out + 2*j + 2*N) = *(Y1out_array + j);
		*(turbo_en_out + 2*j + 1 + 2*N) = *(Y2out_array + j);
	}
	for (j=0; j<M; j++){
		*(turbo_en_out + 2*j + 4*N) = *(W1out_array + 2*j);
		*(turbo_en_out + 2*j + 1 + 4*N) = *(W2out_array + 2*j);
	} 
#endif
	
#if R == 3             /* code rate = 1/2 */
	for (j=0; j<N; j++){
		*(turbo_en_out + 2*j + 2*N) = *(Y1out_array + j);
		*(turbo_en_out + 2*j + 1 + 2*N) = *(Y2out_array + j);
	}
#endif
	
#if R == 4             /* code rate = 2/3 */
	for (j=0; j<M; j++){
		*(turbo_en_out + 2*j + 2*N) = *(Y1out_array + 2*j);
		*(turbo_en_out + 2*j + 1 + 2*N) = *(Y2out_array + 2*j);
	} 
#endif
	
#if R == 5             /* code rate = 3/4 */
	for (j=0; j<M; j++){
		*(turbo_en_out + 2*j + 2*N) = *(Y1out_array + 3*j);
		*(turbo_en_out + 2*j + 1 + 2*N) = *(Y2out_array + 3*j);
	}
#endif 
	
#if R == 6             /* code rate = 4/5 */
	for (j=0; j<M; j++){
		*(turbo_en_out + 2*j + 2*N) = *(Y1out_array + 4*j);
		*(turbo_en_out + 2*j + 1 + 2*N) = *(Y2out_array + 4*j);
	}
#endif 
	
#if R == 7             /* code rate = 6/7 */
	for (j=0; j<M; j++){
		*(turbo_en_out + 2*j + 2*N) = *(Y1out_array + 6*j);
		*(turbo_en_out + 2*j + 1 + 2*N) = *(Y2out_array + 6*j);
	}
#endif
	
	free(pinfo_array);
	free(Y1out_array);
	free(Y2out_array);
	free(W1out_array);
	free(W2out_array);
}

void nbrsc_encd(int g[4][K], int *in_array, int *y_array, int *w_array, int s0[K-1], int sN[K-1]){
	int j;                 /* loop variable */
	int t;					 /* bit time */
	int *a_array,*b_array;   /* pointer to input data arrays */
	int shift_reg[K];        /* the 1st RSC encoder shift register */
	int y,w;				 /* feedback bit and output bits */
	
	a_array = malloc( N * sizeof(int) ); /* allocate space for A input data array */
	if (a_array == NULL) {
		printf("\n nbrsc_encd:  error allocating A array, aborting!");
		exit(1);
	}
	
	b_array = malloc( N * sizeof(int) ); /* allocate space for B input data array */
	if (b_array == NULL) {
		printf("\n nbrsc_encd:  error allocating B array, aborting!");
		exit(1);
	}
	
	for (t=0; t<N; t++){     /* read in the data and store it in the array */
		*(a_array+t) = *(in_array+2*t);
		*(b_array+t) = *(in_array+2*t+1);
	}
	
	shift_reg[0] = 0;        /* Initialize the shift register */
	for (j=0; j<K-1; j++)
		shift_reg[j+1] = s0[j];
	
	for (t=0; t<N; t++){     /* Start the encoding process */
		
		shift_reg[0] = *(a_array + t) ^ *(b_array + t);
		y = 0;
		w = 0;
		
		for (j=1; j<K; j++){   /* get feedback bit */
			shift_reg[0] ^= (shift_reg[j] & g[0][j]);
		}
		
		
		for (j=0; j<K; j++){
			y ^= (shift_reg[j] & g[1][j]);
			w ^= (shift_reg[j] & g[2][j]);
		}
		
		for (j=K-2; j>=0; j--){
			shift_reg[j+1] = shift_reg[j] ^ ( *(b_array+t) & g[3][j] );
		}
		
		*(y_array+t) = y;
		*(w_array+t) = w;
	}
	
	for (j=0; j<(K-1); j++)
		sN[j] = shift_reg[j+1];
	
	free (a_array);         /* free the dynamically allocated array */
	free (b_array);         /* free the dynamically allocated array */
}

/******************************************/
/* Circulation state correspondence table */
/******************************************/
int start_end_state(int sNdeci){
	int s0deci;
	
	s0deci = 0;
	
#if (N%7 == 1)          /* Frame size N modulo 7 = 1 */
	if (sNdeci == 0)
		s0deci = 0;
	else if (sNdeci == 1)
		s0deci = 6;
	else if (sNdeci == 2)
		s0deci = 4;
	else if (sNdeci == 3)
		s0deci = 2;
	else if (sNdeci == 4)
		s0deci = 7;
	else if (sNdeci == 5)
		s0deci = 1;
	else if (sNdeci == 6)
		s0deci = 3;
	else if (sNdeci == 7)
		s0deci = 5;
#endif
	
#if (N%7 == 2)          /* Frame size N modulo 7 = 2 */
	if (sNdeci == 0)
		s0deci = 0;
	else if (sNdeci == 1)
		s0deci = 3;
	else if (sNdeci == 2)
		s0deci = 7;
	else if (sNdeci == 3)
		s0deci = 4;
	else if (sNdeci == 4)
		s0deci = 5;
	else if (sNdeci == 5)
		s0deci = 6;
	else if (sNdeci == 6)
		s0deci = 2;
	else if (sNdeci == 7)
		s0deci = 1;
#endif
	
#if (N%7 == 3)          /* Frame size N modulo 7 = 3 */
	if (sNdeci == 0)
		s0deci = 0;
	else if (sNdeci == 1)
		s0deci = 5;
	else if (sNdeci == 2)
		s0deci = 3;
	else if (sNdeci == 3)
		s0deci = 6;
	else if (sNdeci == 4)
		s0deci = 2;
	else if (sNdeci == 5)
		s0deci = 7;
	else if (sNdeci == 6)
		s0deci = 1;
	else if (sNdeci == 7)
		s0deci = 4;
#endif
	
#if (N%7 == 4)          /* Frame size N modulo 7 = 4 */
	if (sNdeci == 0)
		s0deci = 0;
	else if (sNdeci == 1)
		s0deci = 4;
	else if (sNdeci == 2)
		s0deci = 1;
	else if (sNdeci == 3)
		s0deci = 5;
	else if (sNdeci == 4)
		s0deci = 6;
	else if (sNdeci == 5)
		s0deci = 2;
	else if (sNdeci == 6)
		s0deci = 7;
	else if (sNdeci == 7)
		s0deci = 3;
#endif
	
#if (N%7 == 5)          /* Frame size N modulo 7 = 5 */
	if (sNdeci == 0)
		s0deci = 0;
	else if (sNdeci == 1)
		s0deci = 2;
	else if (sNdeci == 2)
		s0deci = 5;
	else if (sNdeci == 3)
		s0deci = 7;
	else if (sNdeci == 4)
		s0deci = 1;
	else if (sNdeci == 5)
		s0deci = 3;
	else if (sNdeci == 6)
		s0deci = 4;
	else if (sNdeci == 7)
		s0deci = 6;
#endif
	
#if (N%7 == 6)          /* Frame size N modulo 7 = 6 */
	if (sNdeci == 0)
		s0deci = 0;
	else if (sNdeci == 1)
		s0deci = 7;
	else if (sNdeci == 2)
		s0deci = 6;
	else if (sNdeci == 3)
		s0deci = 1;
	else if (sNdeci == 4)
		s0deci = 3;
	else if (sNdeci == 5)
		s0deci = 4;
	else if (sNdeci == 6)
		s0deci = 5;
	else if (sNdeci == 7)
		s0deci = 2;
#endif
	
	return(s0deci);
}

/********************************************/
/* QPSK simulator for non-binary turbo code */
/********************************************/
void gngauss(float mean, float sigma, int encd_block_len, float *gauss1, float *gauss2);
void addnoise(float eb_ovr_n0, int encd_block_len, float rate, int *in_array, float *out_array)
{
	int t;
	float mean, eb, sn_ratio, sigma, signal_1, signal_2;
	float *gauss1, *gauss2;
	
	mean = 0;
	eb = 1;
	sn_ratio = (float) pow(10, ( eb_ovr_n0 / 10) );
	sigma =  (float) sqrt (eb / (2 * sn_ratio * rate) );
	
	gauss1 = malloc( encd_block_len * sizeof(float) );
	if (gauss1 == NULL) {
		printf("\n addnoise:  error allocating cos I noise array, aborting!");
		exit(1);
	}
	
	gauss2 = malloc( encd_block_len * sizeof(float) );
	if (gauss2 == NULL) {
		printf("\n addnoise:  error allocating sin Q noise array, aborting!");
		exit(1);
	}
	
	gngauss(mean, sigma, encd_block_len, gauss1, gauss2);
	
	for (t = 0; t < encd_block_len; t++) {       /* transform the data from 0/1 to +1/-1 and add noise */
		signal_1 = 1 - 2 * *( in_array + 2*t );
		*( out_array + 2*t ) = signal_1 + *(gauss1+t);
		signal_2 = 1 - 2 * *( in_array + 2*t + 1 );
		*( out_array + 2*t + 1) = signal_2 + *(gauss2+t);
	}
	
	free(gauss1);
	free(gauss2);
}

void gngauss(float mean, float sigma, int encd_block_len, float *gauss1, float *gauss2) 
{ 
	int i;
	double u, r;                         /* uniform and Rayleigh random variables */
	
	for (i=0; i<encd_block_len; i++){
		u = (double)rand() / RAND_MAX;     /* uniformly distributed random number u between 0 and 1 - 1E-9*/
		if (u == 1.0) u = 0.999999999;
		
		r = sigma * sqrt( 2.0 * log( 1.0 / (1.0 - u) ) ); /* generate a Rayleigh-distributed random number r using u */
		
		u = (double)rand() / RAND_MAX;             /* generate another uniformly-distributed random number u as before*/
		if (u == 1.0) u = 0.999999999;
		
		*(gauss1+i) = mean + r * cos(2 * PI * u);  /* generate two Gaussian-distributed random number using r and u */
		*(gauss2+i) = mean + r * sin(2 * PI * u);
	} 
}

/**************************************************/
/* Serial to parallel demultiplex at the receiver */
/**************************************************/
void de_multiplex(int msg_length, int M, int *permute_index, float *in_array, float *sys_array,
				  float *pinfo_array, float *parity1_array, float *parity2_array)
{
	int i,j;                    /* loop variables */
	float permu[2][N]={0.0};    /* make a matrix store systematic info array */
	float yy[2][N]={0.0};       /* make matrix to store first parity info of encoder1&2 temporarily */
	float ww[2][N]={0.0};       /* make matrix to store second parity info of encoder1&2 temporarily */
	float temp_data;            /* temporary value */
	
	for (j=0; j<msg_length; j++){       /* get systematic information array */
		*(sys_array + j) = *(in_array + j);
		*(pinfo_array + j) = *(in_array + j);
	}
	
	for (j=0; j<N; j++){                /* turbo code permutation Level one */
		if  (j%2 == 0){
			temp_data = *(pinfo_array + 2*j);
			*(pinfo_array + 2*j) = *(pinfo_array + 2*j + 1);
			*(pinfo_array + 2*j + 1) = temp_data;
		}
	}
	
	for (i=0; i<2; i++){                /* turbo code permutation Level two */
		for (j=0; j<N; j++)
			permu[i][j] = *(pinfo_array + 2*j + i);
	}
	
	for (i=0; i<2; i++){
		for (j=0; j<N; j++)
			*(pinfo_array + 2*j + i) = permu[i][*(permute_index + j)];
	}
	
#if R == 1                          /* code rate = 1/3 */
	for (j=0; j<N; j++){
		for (i=0; i<2; i++){
			yy[i][j] = *(in_array + 2*N + 2*j + i);
			ww[i][j] = *(in_array + 4*N + 2*j + i);
		}
	}
#endif
	
#if R == 2                          /* code rate = 2/5 */
	for (j=0; j<N; j++){
		for (i=0; i<2; i++)
			yy[i][j] = *(in_array + 2*N + 2*j + i);
	}
	for (j=0; j<M; j++){
		for (i=0; i<2; i++){
			ww[i][2*j] = *(in_array + 4*N + 2*j + i);
			ww[i][2*j+1] = 0;
		}
	}
#endif
	
#if R == 3                          /* code rate = 1/2 */
	for (j=0; j<N; j++){
		for (i=0; i<2; i++)
			yy[i][j] = *(in_array + 2*N + 2*j + i);
	}
#endif
	
#if R == 4                          /* code rate = 2/3 */
	for (j=0; j<M; j++){
		for (i=0; i<2; i++){
			yy[i][2*j] = *(in_array + 2*N + 2*j + i);
			yy[i][2*j+1] = 0;
		}
	}
#endif
	
#if R == 5                          /* code rate = 3/4 */
	if (N%3 == 0){         /* N modulo 3 = 1, M=N/3 */
		for (j=0; j<M; j++){
			for (i=0; i<2; i++)
				yy[i][3*j] = *(in_array + 2*N + 2*j + i);
			for (i=1; i<3; i++)
				yy[i-1][3*j+i] = 0;
		}
	}
	else if (N%3 == 1){    /* N modulo 3 = 1, M=(N-4)/3+2 */
		for (j=0; j<(M-1); j++){
			for (i=0; i<2; i++)
				yy[i][3*j] = *(in_array + 2*N + 2*j + i);
			for (i=1; i<3; i++) 
				yy[i-1][3*j+i] = 0;
		}
		yy[0][N-1] = *(in_array + 2*N + 2*M - 2);
		yy[1][N-1] = *(in_array + 2*N + 2*M - 1);
	}
	else if (N%3 == 2){    /* N modulo 3 = 2, M=(N-8)/3+3 */
		for (j=0; j<(M-1); j++){
			for (i=0; i<2; i++)
				yy[i][3*j] = *(in_array + 2*N + 2*j + i);
			for (i=1; i<3; i++)
				yy[i-1][3*j+i] = 0;
		}
		yy[0][N-2] = *(in_array + 2*N + 2*M - 2);
		yy[0][N-1] = 0;
		yy[1][N-2] = *(in_array + 2*N + 2*M - 1);
		yy[1][N-1] = 0;
	}
#endif
	
#if R == 6                         /* code rate = 4/5 */
	for (j=0; j<M; j++){
		for (i=0; i<2; i++)
			yy[i][4*j] = *(in_array + 2*N + 2*j + i);
		for (i=1; i<4; i++){
			yy[0][4*j+i] = 0;
			yy[1][4*j+i] = 0;
		}
	}
#endif
	
#if R == 7                        /* code rate = 6/7 */
	if (N%3 == 0){         /* N modulo 3 = 1, M=N/6 */
		for (j=0; j<M; j++){
			for (i=0; i<2; i++)
				yy[i][6*j] = *(in_array + 2*N + 2*j + i);
			for (i=1; i<6; i++){
				yy[0][6*j+i] = 0;
				yy[1][6*j+i] = 0;
			}
		}
	}
	else if (N%3 == 1){    /* N modulo 3 = 1, M=(N-4)/6+1 */
		for (j=0; j<(M-1); j++){
			for (i=0; i<2; i++)
				yy[i][6*j] = *(in_array + 2*N + 2*j + i);
			for (i=1; i<6; i++){
				yy[0][6*j + i] = 0;
				yy[1][6*j + i] = 0;
			}
		}
		yy[0][N-4] = *(in_array + 2*N + 2*M - 2);
		for (j=3; j>0; j--)
			yy[0][N-j] = 0;
		yy[1][N-4] = *(in_array + 2*N + 2*M - 1);
		for (j=3; j>0; j--)
			yy[1][N-j] = 0;
	}
	else if (N%3 == 2){    /* N modulo 3 = 2, M=(N-8)/6+2 */
		for (j=0; j<(M-1); j++){
			for (i=0; i<2; i++)
				yy[i][6*j] = *(in_array + 2*N + 2*j + i);
			for (i=1; i<6; i++){
				yy[0][6*j + i] = 0;
				yy[1][6*j + i] = 0;
			}
		}
		yy[0][N-2] = *(in_array + 2*N + 2*M - 2);
		yy[0][N-1] = 0;
		yy[1][N-2] = *(in_array + 2*N + 2*M - 1);
		yy[1][N-1] = 0;
	}
#endif
	
	if (R >= 3){
		for (j=0; j<N; j++){
			for (i=0; i<2; i++)
				ww[i][j] = 0;
		}
	}
	
	for (j=0; j<N; j++){
		*(parity1_array + 2*j) = yy[0][j];
		*(parity1_array + 2*j + 1) = ww[0][j];
		*(parity2_array + 2*j) = yy[1][j];
		*(parity2_array + 2*j + 1) = ww[1][j];
	}
}

/*******************************************/
/* Non-binary turbo code permutation table */
/*******************************************/
void permutation(int *permute_index)
{
	int j;         /* loop variables */
	
#if N == 48
	int p[4] = {11,24,0,24};
#endif
	
#if N == 64
	int p[4] = {7,34,32,2};
#endif
	
#if N == 212
	int p[4] = {13,106,108,2};
#endif
	
#if N == 220
	int p[4] = {23,112,4,116};
#endif
	
#if N == 228
	int p[4] = {17,116,72,188};
#endif
	
#if N == 424
	int p[4] = {11,6,8,2};
#endif
	
#if N == 432
	int p[4] = {13,0,4,8};
#endif
	
#if N == 440
	int p[4] = {13,10,4,2};
#endif
	
#if N == 752
	int p[4] = {19,376,224,600};
#endif
	
#if N == 848
	int p[4] = {19,2,16,6};
#endif
	
#if N == 856
	int p[4] = {19,428,224,652};
#endif
	
#if N == 864
	int p[4] = {19,2,16,6};
#endif
	
	for (j=0; j<N; j++){
		if (j%4 == 0) *(permute_index + j) = (p[0]*j + 1) % N;
		else if (j%4 == 1) *(permute_index + j) = (p[0]*j + N/2 + p[1] + 1) % N;
		else if (j%4 == 2) *(permute_index + j) = (p[0]*j + p[2] + 1) % N;
		else if (j%4 == 3) *(permute_index + j) = (p[0]*j + N/2 + p[3] + 1) % N;
	}
}

/****************************************/
/* Turbo code decoding with Max_log_map */
/*
/* int nextoutput[NUMSTATES][NUMSTATES]		- maps current/next states to next output */
/* int nextstate[NUMSTATES][SYMBOLS]		- for current state, gives next for given input */
/* int preoutput[NUMSTATES][NUMSTATES]		- maps previous/current states to previous output */
/* int prestate[NUMSTATES][SYMBOLS]			- gives conv. encoder output */
/****************************************/
float max4(float x, float y, float z, float w);
float max8(float x, float y, float z, float w, float A, float B, float C, float D);

void maxlogmap(float *rec_s, float *rec_p, float *L_a, float *L_u, float *Alpha0, float *BetaN,
			   int nextstate[NUMSTATES][SYMBOLS], int nextoutput[NUMSTATES][NUMSTATES][2],
			   int prestate[NUMSTATES][SYMBOLS], int preoutput[NUMSTATES][NUMSTATES][2], int qpsk_input[SYMBOLS][2])
{
	int i,j,t,s;							/* loop variables */
	
	float gamma[NUMSTATES] = {0.0};         /* the branch transition probabilities */
	float Alpha[N+1][NUMSTATES];
	float Beta[N+1][NUMSTATES];
	
	float Alpha_temp[SYMBOLS] = {0.0};		/* temporary values of Alpha or Beta */
	float Beta_temp[SYMBOLS] = {0.0};
	float tempmax[N] = {0.0};
	float temp0[NUMSTATES] = {0.0};
	float temp1[NUMSTATES] = {0.0};
	float temp2[NUMSTATES] = {0.0};
	float temp3[NUMSTATES] = {0.0};
	float max8temp0 = {0.0};
	float gamma0[SYMBOLS] = {0.0};
	float p[N][SYMBOLS] = {0.0};
	
	for (i = 0; i < NUMSTATES; i++){        /* initialize Alpha and Beta data array */
		Alpha[0][i] = *(Alpha0 + i);
		Beta[N][i] = *(BetaN + i);
	}
	
	for (j=0; j<N; j++){
		p[j][0] = - max4(0,*(L_a+j),*(L_a+j+N),*(L_a+j+2*N));
		p[j][1] = *(L_a+j) + p[j][0];
		p[j][2] = *(L_a+j+N) + p[j][0];
		p[j][3] = *(L_a+j+2*N) + p[j][0];
	}
	
	for (t=1; t<(N+1); t++) {        /* trace forward, compute Alpha */
		for (j=0; j<NUMSTATES; j++) {
			for (i=0; i<SYMBOLS; i++){
				s = prestate[j][i];
				gamma[s] = *(rec_s+2*t-2)*qpsk_input[i][0] + *(rec_s+2*t-1)*qpsk_input[i][1]
				+ *(rec_p+2*t-2)*preoutput[j][s][0] + *(rec_p+2*t-1)*preoutput[j][s][1] + p[t-1][i];
				
				Alpha_temp[i] = gamma[s] + Alpha[t-1][s];
			}
			Alpha[t][j] = max4(Alpha_temp[0], Alpha_temp[1], Alpha_temp[2], Alpha_temp[3]);
		}
		tempmax[t-1] = max8(Alpha[t][0],Alpha[t][1],Alpha[t][2],Alpha[t][3],Alpha[t][4],Alpha[t][5],Alpha[t][6],Alpha[t][7]);
		
		for (j=0; j<NUMSTATES; j++)    /* to be sure that Alpha do not overflow */
			Alpha[t][j] = Alpha[t][j] - tempmax[t-1];
	}
	
	for (t=N-1; t>=0; t--) {         /* trace backward, compute Beta */
		for (j=0; j<NUMSTATES; j++) {
			for (i=0; i<SYMBOLS; i++){
				s = nextstate[j][i];
				gamma[s] = *(rec_s+2*t)*qpsk_input[i][0] + *(rec_s+2*t+1)*qpsk_input[i][1]
				+ *(rec_p+2*t)*nextoutput[j][s][0] + *(rec_p+2*t+1)*nextoutput[j][s][1] + p[t][i];
				
				Beta_temp[i] = gamma[s] + Beta[t+1][s];
			}
			Beta[t][j] = max4(Beta_temp[0], Beta_temp[1], Beta_temp[2], Beta_temp[3]);
		}
		
		for (j=0; j<NUMSTATES; j++)    /* to be sure that Beta do not overflow */
			Beta[t][j] = Beta[t][j] - tempmax[t];
	}
	
	for (t=0; t<N; t++) {            /* compute the soft output, log_likelihood ratio */
		for (j=0; j<NUMSTATES; j++) {
			for (i=0; i<SYMBOLS; i++){
				s = prestate[j][i];
				gamma0[i] = *(rec_s+2*t)*qpsk_input[i][0] + *(rec_s+2*t+1)*qpsk_input[i][1]
				+ *(rec_p+2*t)*preoutput[j][s][0] + *(rec_p+2*t+1)*preoutput[j][s][1] + p[t][i];
			}
			
			temp0[j] = gamma0[0] + Alpha[t][prestate[j][0]] + Beta[t+1][j]; 
			
			temp1[j] = gamma0[1] + Alpha[t][prestate[j][1]] + Beta[t+1][j];
			
			temp2[j] = gamma0[2] + Alpha[t][prestate[j][2]] + Beta[t+1][j];
			
			temp3[j] = gamma0[3] + Alpha[t][prestate[j][3]] + Beta[t+1][j];
		}
		max8temp0 = max8(temp0[0],temp0[1],temp0[2],temp0[3],temp0[4],temp0[5],temp0[6],temp0[7]);
		
		*(L_u+t) = max8(temp1[0],temp1[1],temp1[2],temp1[3],temp1[4],temp1[5],temp1[6],temp1[7]) - max8temp0;
		
		*(L_u+t+N) = max8(temp2[0],temp2[1],temp2[2],temp2[3],temp2[4],temp2[5],temp2[6],temp2[7]) - max8temp0;
		
		*(L_u+t+2*N) = max8(temp3[0],temp3[1],temp3[2],temp3[3],temp3[4],temp3[5],temp3[6],temp3[7]) - max8temp0;
	}
	
	for (i = 0; i < NUMSTATES; i++){
		*(Alpha0 + i) = Alpha[N][i];
		*(BetaN + i) = Beta[0][i];
	}
}

float max4(float x, float y, float z, float w)
{
	float maxmum;
	
	maxmum = x>y?x:y;
	maxmum = maxmum>z?maxmum:z;
	maxmum = maxmum>w?maxmum:w;
	
	return(maxmum);
}

float max8(float x, float y, float z, float w, float A, float B, float C, float D)
{
	float maxmum;
	
	maxmum = x>y?x:y;
	maxmum = maxmum>z?maxmum:z;
	maxmum = maxmum>w?maxmum:w;
	maxmum = maxmum>A?maxmum:A;
	maxmum = maxmum>B?maxmum:B;
	maxmum = maxmum>C?maxmum:C;
	maxmum = maxmum>D?maxmum:D;
	
	return(maxmum);
}

/*************************************************/
/* Set up the trellis for a given code generator */
/*************************************************/
void deci2bin(int d, int size, int *b);
int bin2deci(int *b, int size);
int nxt_stat(int g[4][K], int current_state, int input, int *memory_contents);
void qpskmodulate(int *array, int size);

void trellis(int g[4][K], int nextstate[NUMSTATES][SYMBOLS], int nextoutput[NUMSTATES][NUMSTATES][2],
			 int prestate[NUMSTATES][SYMBOLS], int preoutput[NUMSTATES][NUMSTATES][2])
{
	int i,j,l;                   /* loop variables */
	int memory_contents[K];      /* feedback bit + conv. encoder sr */
	int encd_output[2];          /* store trial encoder output */
	int next_state;
	
	for (i = 0; i < NUMSTATES; i++) {     /* initialize data structures */
		for (j = 0; j < NUMSTATES; j++){
			preoutput[i][j][0] = 0;
			preoutput[i][j][1] = 0;
			nextoutput[i][j][0] = 0;
			nextoutput[i][j][1] = 0;
		}
		
		for (j = 0; j < SYMBOLS; j++){
			prestate[i][j] = 0;
			nextstate[i][j] = 0;
		}
	}
	
	for (j = 0; j < NUMSTATES; j++) {     /* generate the matrices */
		for (l = 0; l < K; l++) {
			next_state = nxt_stat(g, j, l, memory_contents);
			
			encd_output[0] = 0;
			encd_output[1] = 0;
			for (i = 0; i < K; i++){
				encd_output[0] ^= memory_contents[i] & g[1][i];
				encd_output[1] ^= memory_contents[i] & g[2][i];
			}
			
			qpskmodulate(encd_output, 2);	/* Modulate the bits, so they match the channel modulation */
			
			nextstate[j][l] = next_state;	/* Next state, given current state and input */
			prestate[next_state][l] = j;	/* Previous state, given current state and input */
			
			nextoutput[j][next_state][0] = encd_output[0];	/* The first bit of the next output */			
			nextoutput[j][next_state][1] = encd_output[1];	/* The second bit of the next output */
			
			preoutput[next_state][j][0] = encd_output[0];	/* The first bit of the previous output */		
			preoutput[next_state][j][1] = encd_output[1];	/* The second bit of the previous output */
		}  /* end of l for loop */
	}  /* end of j for loop */
}

/* ************************************************************************** */
int nxt_stat(int g[4][K], int current_state, int input, int *memory_contents) {
	int binary_state[K - 1];              /* binary value of current state */
	int binary_input[2];                  /* binary value of input */
	int next_state_binary[K - 1];         /* binary value of next state */
	int next_state;                       /* decimal value of next state */
	int feedback;                         /* feedback bit */
	int i;                                /* loop variable */
	
	deci2bin(current_state, K - 1, binary_state);	/* convert the decimal value of the current state number to binary */
	deci2bin(input, 2, binary_input);				/* convert the decimal value of the input to binary */
	
	/* Get the feedback bit */
	feedback = binary_input[0] ^ binary_input[1];
	for (i=0; i<K-1; i++){
		feedback ^= binary_state[i] & g[0][i+1];	
	}
	
	/* Given the input and current state number, compute the next state number */
	next_state_binary[0] = feedback;
	for (i = 1; i < K - 1; i++){
		next_state_binary[i] = binary_state[i - 1] ^ (binary_input[1]  & g[3][i]);
	}
		
	/* Convert state number to decimal */
	next_state = bin2deci(next_state_binary, K - 1);
	
	/* Load the memory contents with the current state, for use by the trellis method*/
	memory_contents[0] = feedback;
	for (i = 0; i < K-1; i++){
		memory_contents[i+1] = binary_state[i];
	}
	
	return(next_state);
}

void deci2bin(int d, int size, int *b) {	/* Converts a decimal number to a binary number */
	int i;
	
	for(i = 0; i < size; i++)
		b[i] = 0;
	
	b[size - 1] = d & 0x01;
	
	for (i = size - 2; i >= 0; i--) {
		d = d >> 1;
		b[i] = d & 0x01;
	}
}

int bin2deci(int *b, int size) {			/* Converts a binary number to decimal number */
	int i, d;
	
	d = 0;
	
	for (i = 0; i < size; i++)
		d = d + (int)pow(2, size - i - 1) * b[i];
	
	return(d);
}

void qpskmodulate(int *input, int size){	/* Converts 0/1 to +1/-1, as in the qpsk modulation */
	int l;
	for (l=0; l<size; l++){
		input[l] = 1 - 2 * input[l];
	}
}
